const ArtefactServer = require('artefact-server')
const { artefactEncryptionKey } = require('artefact-store')
const path = require('path')
const Obz = require('obz')
const defer = require('pull-defer')
const toPull = require('stream-to-pull-stream')
const { v4: uuid } = require('uuid') // TODO check which version to use!

const Prune = require('./prune')
const AutoPrune = require('./auto-prune')

module.exports = {
  name: 'hyperBlobs',
  version: require('./package.json').version,
  manifest: {
    onReady: 'async',
    add: 'sink',
    driveAddress: 'async',
    registerDrive: 'async',
    prune: 'async'
  },
  permissions: {
    anonymous: {
      allow: ['registerDrive']
    }
  },
  init (ssb, config) {
    const {
      port = 26836,
      pataka = false,
      storeOpts = {},
      autoPrune = false
    } = (config.hyperBlobs || {})

    const artefactServerObz = Obz()
    ArtefactServer({
      storage: path.resolve(config.path, 'hyperBlobs'),
      port,
      pataka,
      storeOpts
    })
      .then(_server => artefactServerObz.set(_server))
      .catch(err => { throw err })

    ssb.close.hook(function (close, args) {
      const self = this
      artefactServerObz.once(artefactServer => {
        artefactServer.close()
          .then(() => {
            close.apply(self, args)
          })
      })
    })

    if (!pataka) {
      // TODO have pataka register all the drive addresses it knows with peers who connect!

      const peerAttempts = {
        // peerId: attemptCount
      }
      ssb.on('rpc:connect', remote => {
        if (remote.id === ssb.id) return // could be ssb-client connection
        if (!remote.hyperBlobs) return

        driveAddress((err, address) => {
          if (err) throw err

          peerAttempts[remote.id] = peerAttempts[remote.id] || 0
          if (remote.closed) return reconnect(remote)
          // NOTE: the artefact-store takes a while to get ready
          // the rpc connecting can closes if there's is no action, here we manually start the connection again

          attemptRegister(remote, address)
        })
      })

      function reconnect (remote) {
        if (!ssb.conn) {
          console.error('ssb-hyper-blobs needs ssb-conn to restart closed connections')
          return
        }
        ssb.conn.connect(remote.stream.address)
      }
      function attemptRegister (remote, address) {
        peerAttempts[remote.id] = peerAttempts[remote.id] + 1

        remote.hyperBlobs.registerDrive(address, (err) => {
          if (!err) {
            delete peerAttempts[remote.id]
            return
          }
          if (peerAttempts[remote.id] < 5) {
            console.log(`remote.hyperBlobs.registerDrive failed (attempt ${peerAttempts[remote.id]})`)
            if (remote.closed) return reconnect(remote)
            return setTimeout(() => attemptRegister(remote, address), 2000)
          }

          console.error('hyperBlobs.registerDrive failed. Giving up', err)
        })
      }
    }

    if (autoPrune) artefactServerObz.once(as => AutoPrune(as, ssb))

    // let _driveAddress
    function driveAddress (cb) {
      if (pataka) return cb(new Error('pataka does not have its own driveAddress'))

      artefactServerObz.once(artefactServer => {
        cb(null, toString(artefactServer.store.driveAddress))
      })
    }

    return {
      onReady (cb) {
        artefactServerObz.once((as) => cb(null, as))
      },
      add (cb) {
        const sink = defer.sink()

        const blobId = uuid()
        const readKey = artefactEncryptionKey()
        const opts = {}

        artefactServerObz.once(artefactServer => {
          const writeStream = artefactServer.store.createWriteStream(blobId, readKey, opts)

          const _sink = toPull.sink(writeStream, (err) => {
            if (err) return cb(err)

            artefactServer.store.fileReady(blobId)
              .then(() => {
                cb(null, {
                  driveAddress: toString(artefactServer.store.driveAddress),
                  blobId,
                  readKey: toString(readKey)
                  // fileName....
                })
              })
          })
          sink.resolve(_sink)
        })

        return sink
      },
      driveAddress,
      registerDrive (address, cb) {
        artefactServerObz.once(artefactServer => {
          artefactServer.store.addDrive(address)
            .then(() => cb(null, true))
            .catch(err => cb(err))
        })
      },
      prune (opts, cb) {
        artefactServerObz.once(as => {
          Prune(as)(opts, cb)
        })
      }
    }
  }
}

function toString (buf) {
  if (typeof buf === 'string') return buf
  return buf.toString('hex')
}
